package pps.core.base.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.IoUtil;
import cn.hutool.core.io.file.FileReader;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.sun.mail.imap.IMAPFolder;
import org.apache.commons.lang3.StringUtils;
import pps.cloud.base.service.IBaseWeatherCloudService;
import pps.core.base.entity.BasePhotovoltaicPlantView;
import pps.core.base.entity.PlantPredictedPowerDataEnt;
import pps.core.base.enums.WindDirection;
import pps.core.base.mapper.BasePhotovoltaicPlantViewMapper;
import pps.core.base.mapper.PlantPredictedPowerDataMapper;
import pps.core.base.service.data.DailyData;
import pps.core.common.constant.BusinessConstant;
import xstartup.annotation.XService;
import xstartup.base.XContext;
import xstartup.base.XStartup;
import xstartup.base.util.XCopyUtils;
import xstartup.data.XServiceResult;

import javax.mail.*;
import javax.mail.internet.MimeUtility;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.time.LocalTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 天气数据接收Cloud模块
 *
 * @author ZWT
 * @date 2023/11/13 17:54
 */
@XService
public class BaseWeatherCloudServiceImpl implements IBaseWeatherCloudService {

    /**
     * 天气数据接收Cloud模块--天气数据处理
     *
     * @param context 上下文
     * @return {@link XServiceResult}
     */
    @Override
    public XServiceResult weatherDataProcess(XContext context) {
        String host = XStartup.getCurrent().getProperty("weather.mail.host");
        String username = XStartup.getCurrent().getProperty("weather.mail.username");
        String password = XStartup.getCurrent().getProperty("weather.mail.password");
        String protocol = XStartup.getCurrent().getProperty("weather.mail.protocol");
        //取未读邮件
        Store store = null;
        Folder folder = null;
        try {
            Properties props = System.getProperties();
            props.setProperty("mail.store.protocol", protocol);
            Session session = Session.getDefaultInstance(props, null);
            store = session.getStore(protocol);
            store.connect(host, username, password);
            folder = store.getFolder("INBOX");
            if (folder instanceof IMAPFolder) {
                IMAPFolder imapFolder = (IMAPFolder) folder;
                // javamail中使用id命令有校验checkOpened, 所以要去掉id方法中的checkOpened();
                imapFolder.doCommand(p -> {
                    Map<String, String> gmap = new HashMap<>();
                    gmap.put("GUID", "FUTONG");
                    p.id(gmap);
                    return null;
                });
            }
            if (Objects.nonNull(folder)) {
                folder.open(Folder.READ_WRITE);
            }
            Message[] messages = folder.getMessages();
            if (ArrayUtil.isNotEmpty(messages)) {
                List<Message> messageList = new ArrayList<>(16);
                for (Message message : messages) {
                    boolean isNew = false;
                    for (Flags.Flag flag : message.getFlags().getSystemFlags()) {
                        if (flag == Flags.Flag.SEEN) {
                            isNew = true;
                        }
                    }
                    if (!isNew && StrUtil.endWith(message.getSubject(), "天气预测数据")) {
                        messageList.add(message);
                    }
                }
                if (CollUtil.isEmpty(messageList)) {
                    return null;
                }
                //取电站
                BasePhotovoltaicPlantViewMapper viewMapper = context.getBean(BasePhotovoltaicPlantViewMapper.class);
                List<BasePhotovoltaicPlantView> plantList = viewMapper.selectPlantList();
                if (CollUtil.isEmpty(plantList)) {
                    return null;
                }
                //按照cityCode分组
                Map<String, List<BasePhotovoltaicPlantView>> plantMap = plantList.stream()
                        .collect(Collectors.groupingBy(BasePhotovoltaicPlantView::getCityCode));
                PlantPredictedPowerDataMapper mapper = context.getBean(PlantPredictedPowerDataMapper.class);
                List<String> timeList = Arrays.asList("15:00", "30:00", "45:00");
                //读取附件
//                for (Message message : messageList) {
//                    String tempFilePath = this.saveAttachment(message);
//                    FileReader fileReader = new FileReader(tempFilePath);
//                    String jsonString = fileReader.readString();
//                    JSONObject jsonObject = JSONObject.parseObject(jsonString);
//                    Map<String, List<DailyData>> cityDataMap = new HashMap<>(16);
//                    for (Map.Entry<String, Object> entry : jsonObject.entrySet()) {
//                        String key = entry.getKey();
//                        Object value = entry.getValue();
//                        if (value instanceof JSONArray) {
//                            cityDataMap.put(key, ((JSONArray) value).toJavaList(DailyData.class));
//                        }
//                    }
//
//
//                    FileUtil.del(tempFilePath);
//                }
                FileReader fileReader = new FileReader("D:\\home\\weather\\temp\\receive\\20231114\\1699943400000.json");
                String jsonString = fileReader.readString();
                JSONObject jsonObject = JSONObject.parseObject(jsonString);
                Map<String, List<DailyData>> cityDataMap = new HashMap<>(16);
                for (Map.Entry<String, Object> entry : jsonObject.entrySet()) {
                    String key = entry.getKey();
                    Object value = entry.getValue();
                    if (value instanceof JSONArray) {
                        cityDataMap.put(key, ((JSONArray) value).toJavaList(DailyData.class));
                    }
                }
                for (String cityCode : cityDataMap.keySet()) {
                    if (!plantMap.containsKey(cityCode)) {
                        continue;
                    }
                    List<DailyData> dailyDataList = cityDataMap.get(cityCode);
                    List<BasePhotovoltaicPlantView> plantViewList = plantMap.get(cityCode);
                    //遍历预测数据列表
                    for (DailyData dailyData : dailyDataList) {
                        for (BasePhotovoltaicPlantView plant : plantViewList) {
                            List<PlantPredictedPowerDataEnt> allList = new ArrayList<>();
                            List<PlantPredictedPowerDataEnt> objList = new ArrayList<>();
                            DateTime dateTime = DateTime.of(dailyData.getDateTime());
                            int hours = 0;
                            String dayTime = dateTime.toString("yyyy-MM-dd");
                            String plantId = plant.getId();
                            DateTime date;
                            for (int i = 0; i < 8; i++) {
                                //判断小时数与查到的小时数，如果查到的小时数小，判断为第二天
                                int hour = LocalTime.parse(dailyData.getHourTimeArray().get(i)).getHour();
                                if (hours >= hour) {
                                    dateTime = DateUtil.offsetDay(dateTime, 1);
                                    dayTime = dateTime.toString("yyyy-MM-dd");
                                }
                                hours = hour;
                                PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                        this.int2Str(hours),
                                        new BigDecimal(StrUtil.replace(dailyData.getTemperatureArray().get(i), "℃", "")),
                                        new BigDecimal(StrUtil.replace(dailyData.getWindSpeedArray().get(i), "m/s", "")),
                                        new BigDecimal(WindDirection.findByValue(dailyData.getWindDirectionArray().get(i)).getValue()),
                                        new BigDecimal(StrUtil.replace(dailyData.getPressureArray().get(i), "hPa", "")),
                                        new BigDecimal(StrUtil.replace(dailyData.getHumidityArray().get(i), "%", "")),
                                        String.valueOf(dateTime.year()),
                                        this.int2Str(dateTime.month() + 1),
                                        this.int2Str(dateTime.dayOfMonth()),
                                        BusinessConstant.START_OF_DAY_TIME
                                );
                                // 5.执行SQL
                                mapper.delete(new LambdaQueryWrapper<PlantPredictedPowerDataEnt>()
                                        .eq(PlantPredictedPowerDataEnt::getPlantId, plantId)
                                        .eq(PlantPredictedPowerDataEnt::getYearTime, dao.getYearTime())
                                        .eq(PlantPredictedPowerDataEnt::getMonthTime, dao.getMonthTime())
                                        .eq(PlantPredictedPowerDataEnt::getDayTime, dao.getDayTime())
                                        .eq(PlantPredictedPowerDataEnt::getHourTime, dao.getHourTime())
                                );
                                allList.add(dao);
                                PlantPredictedPowerDataEnt isExit = objList.stream().filter(item ->
                                                StrUtil.equals(item.getYearTime(), dao.getYearTime()) &&
                                                        StrUtil.equals(item.getMonthTime(), dao.getMonthTime()) &&
                                                        StrUtil.equals(item.getDayTime(), dao.getDayTime()) &&
                                                        StrUtil.equals(item.getHourTime(), dao.getHourTime()))
                                        .findFirst()
                                        .orElse(null);
                                if (Objects.isNull(isExit)) {
                                    objList.add(dao);
                                }
                            }
                            // 原站点时间结构为：08:00	11:00	14:00	17:00	20:00	23:00	02:00	05:00
                            // 两条数据中间差了两个小时，使用上下两条数据推算出中间差掉的两个小时
                            for (int i = 0; i < objList.size() - 1; i++) {
                                for (int y = 1; y < 3; y++) {
                                    date = DateUtil.parseDate(objList.get(i).getDataDate());
                                    String hourString;
                                    Integer hourTime = Integer.valueOf(objList.get(i).getHourTime());
                                    if (hourTime.equals(23)) {
                                        date = DateUtil.offsetDay(date, 1);
                                        hourString = "0" + (y - 1);
                                    } else {
                                        if ((hourTime + y) < 10) {
                                            hourString = "0" + (hourTime + y);
                                        } else {
                                            hourString = String.valueOf(hourTime + y);
                                        }
                                    }
                                    dayTime = date.toString("yyyy-MM-dd");
                                    int nextIndex = i + 1;
                                    PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                            hourString,
                                            this.compute(objList.get(i).getTemperature(), objList.get(nextIndex).getTemperature(), y),
                                            this.compute(objList.get(i).getWindSpeed(), objList.get(nextIndex).getWindSpeed(), y),
                                            objList.get(i).getWindDirection(),
                                            this.compute(objList.get(i).getPressure(), objList.get(nextIndex).getPressure(), y),
                                            this.compute(objList.get(i).getHumidity(), objList.get(nextIndex).getHumidity(), y),
                                            String.valueOf(date.year()),
                                            this.int2Str(date.month() + 1),
                                            this.int2Str(date.dayOfMonth()),
                                            BusinessConstant.START_OF_DAY_TIME
                                    );
                                    // 5.执行SQL
                                    mapper.delete(new LambdaQueryWrapper<PlantPredictedPowerDataEnt>()
                                            .eq(PlantPredictedPowerDataEnt::getPlantId, plantId)
                                            .eq(PlantPredictedPowerDataEnt::getYearTime, dao.getYearTime())
                                            .eq(PlantPredictedPowerDataEnt::getMonthTime, dao.getMonthTime())
                                            .eq(PlantPredictedPowerDataEnt::getDayTime, dao.getDayTime())
                                            .eq(PlantPredictedPowerDataEnt::getHourTime, dao.getHourTime())
                                    );
                                    allList.add(dao);
                                }
                            }
                            //入库
                            List<PlantPredictedPowerDataEnt> batchList = new ArrayList<>();
                            List<PlantPredictedPowerDataEnt> sortList = allList.stream()
                                    .sorted(Comparator.comparing(PlantPredictedPowerDataEnt::getDataDate))
                                    .collect(Collectors.toList());
                            for (int i = 0; i < sortList.size(); i++) {
                                PlantPredictedPowerDataEnt item = sortList.get(i);
                                batchList.add(item);
                                if (i < sortList.size() - 1) {
                                    for (int y = 0; y < timeList.size(); y++) {
                                        PlantPredictedPowerDataEnt dao = XCopyUtils.copyNewObject(item);
                                        dao.setMinTime(timeList.get(y));
                                        dao.setDataDate(dao.getYearTime() + "-" + dao.getMonthTime() + "-" + dao.getDayTime() + " " + dao.getHourTime() + ":" + dao.getMinTime());
                                        int nextIndex = i + 1;
                                        BigDecimal temperature_1 = sortList.get(i).getTemperature() == null ? new BigDecimal(0) : sortList.get(i).getTemperature();
                                        BigDecimal temperature_2 = sortList.get(nextIndex).getTemperature() == null ? new BigDecimal(0) : sortList.get(nextIndex).getTemperature();
                                        dao.setTemperature(temperature_1.add(temperature_2.subtract(temperature_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal humidity_1 = sortList.get(i).getHumidity() == null ? new BigDecimal(0) : sortList.get(i).getHumidity();
                                        BigDecimal humidity_2 = sortList.get(nextIndex).getHumidity() == null ? new BigDecimal(0) : sortList.get(nextIndex).getHumidity();
                                        dao.setHumidity(humidity_1.add(humidity_2.subtract(humidity_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal windSpeed_1 = sortList.get(i).getWindSpeed() == null ? new BigDecimal(0) : sortList.get(i).getWindSpeed();
                                        BigDecimal windSpeed_2 = sortList.get(nextIndex).getWindSpeed() == null ? new BigDecimal(0) : sortList.get(nextIndex).getWindSpeed();
                                        dao.setWindSpeed(windSpeed_1.add(windSpeed_2.subtract(windSpeed_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal pressure_1 = sortList.get(i).getPressure() == null ? new BigDecimal(0) : sortList.get(i).getPressure();
                                        BigDecimal pressure_2 = sortList.get(nextIndex).getPressure() == null ? new BigDecimal(0) : sortList.get(nextIndex).getPressure();
                                        dao.setPressure(pressure_1.add(pressure_2.subtract(pressure_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal planeIrradiance_1 = sortList.get(i).getPlaneIrradiance() == null ? new BigDecimal(0) : sortList.get(i).getPlaneIrradiance();
                                        BigDecimal planeIrradiance_2 = sortList.get(nextIndex).getPlaneIrradiance() == null ? new BigDecimal(0) : sortList.get(nextIndex).getPlaneIrradiance();
                                        dao.setPlaneIrradiance(planeIrradiance_1.add(planeIrradiance_2.subtract(planeIrradiance_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal horizontalIrradiance_1 = sortList.get(i).getHorizontalIrradiance() == null ? new BigDecimal(0) : sortList.get(i).getHorizontalIrradiance();
                                        BigDecimal horizontalIrradiance_2 = sortList.get(nextIndex).getHorizontalIrradiance() == null ? new BigDecimal(0) : sortList.get(nextIndex).getHorizontalIrradiance();
                                        dao.setHorizontalIrradiance(horizontalIrradiance_1.add(horizontalIrradiance_2.subtract(horizontalIrradiance_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        BigDecimal power_1 = sortList.get(i).getPower() == null ? new BigDecimal(0) : sortList.get(i).getPower();
                                        BigDecimal power_2 = sortList.get(nextIndex).getPower() == null ? new BigDecimal(0) : sortList.get(nextIndex).getPower();
                                        dao.setPower(power_1.add(power_2.subtract(power_1).multiply(new BigDecimal(0.3 * (y + 1)))));
                                        batchList.add(dao);
                                    }
                                }

                            }


                        }


                    }
                }
            }
        } catch (MessagingException e) {
            throw new RuntimeException(e);
        } finally {
            IoUtil.close(folder);
            IoUtil.close(store);
        }
        return XServiceResult.OK;
    }

    /**
     * 获取电厂预测功率ent
     *
     * @param plantId       电站ID
     * @param hourTime      小时
     * @param temperature   温度
     * @param windSpeed     风速
     * @param windDirection 风向
     * @param pressure      压力
     * @param humidity      湿度
     * @param yearTime      年时间
     * @param monthTime     月份时间
     * @param dayTime       白天
     * @param minTime       分钟
     * @return {@link PlantPredictedPowerDataEnt}
     */
    private PlantPredictedPowerDataEnt getPlantPredictedPowerEnt(String plantId, String hourTime, BigDecimal temperature, BigDecimal windSpeed, BigDecimal windDirection,
                                                                 BigDecimal pressure, BigDecimal humidity, String yearTime, String monthTime, String dayTime, String minTime) {
        PlantPredictedPowerDataEnt dao = new PlantPredictedPowerDataEnt();
        dao.setCreateTime(new Date());
        dao.setPlantId(plantId);
        dao.setHourTime(hourTime);
        dao.setTemperature(temperature);
        dao.setWindSpeed(windSpeed);
        dao.setWindDirection(windDirection);
        dao.setPressure(pressure);
        dao.setHumidity(humidity);
        dao.setYearTime(yearTime);
        dao.setMonthTime(monthTime);
        dao.setDayTime(dayTime);
        dao.setMinTime(minTime);
        dao.setDataDate(new StringBuilder(dao.getYearTime())
                .append('-')
                .append(dao.getMonthTime())
                .append('-')
                .append(dao.getDayTime())
                .append(' ')
                .append(dao.getHourTime())
                .append(':')
                .append(dao.getMinTime()).toString());
        return dao;
    }


    private BigDecimal compute(BigDecimal a, BigDecimal b, int y) {
        return a.add(b.subtract(a).multiply(new BigDecimal(0.3 * y)));
    }

    /**
     * int转String
     *
     * @param num 数字
     * @return {@link String}
     */
    private String int2Str(int num) {
        return num < 10 ? "0" + num : String.valueOf(num);
    }


    /**
     * 保存附件
     *
     * @param part 部分
     * @return {@link String}
     * @throws MessagingException 消息传递异常
     * @throws IOException        IOException
     */
    private String saveAttachment(Part part) throws MessagingException, IOException {
        String fileName;
        String filePath = null;
        if (part.isMimeType("multipart/*")) {
            Multipart mp = (Multipart) part.getContent();
            BodyPart mPart;
            for (int i = 0; i < mp.getCount(); i++) {
                mPart = mp.getBodyPart(i);
                String disposition = mPart.getDisposition();
                if (Part.ATTACHMENT.equals(disposition) || Part.INLINE.equals(disposition)) {
                    fileName = mPart.getFileName();
                    if (StrUtil.isNotEmpty(fileName) && StringUtils.lowerCase(fileName).indexOf("GB2312") != -1) {
                        fileName = MimeUtility.decodeText(fileName);
                    }
                    fileName = StrUtil.replace(fileName, "\\?|=", "");
                    filePath = this.saveTempFile(fileName, mPart.getInputStream());
                } else if (mPart.isMimeType("multipart/*")) {
                    this.saveAttachment(mPart);
                } else {
                    fileName = mPart.getFileName();
                    if (StrUtil.isNotEmpty(fileName) && (StringUtils.lowerCase(fileName).indexOf("GB2312") != -1)) {
                        fileName = MimeUtility.decodeText(fileName);
                        filePath = this.saveTempFile(fileName, mPart.getInputStream());
                    }
                }
            }
        } else if (part.isMimeType("message/rfc822")) {
            this.saveAttachment((Part) part.getContent());
        }
        return filePath;
    }

    /**
     * 保存临时文件
     *
     * @param fileName 文件名
     * @param in       在里面
     * @return {@link String}
     */
    private String saveTempFile(String fileName, InputStream in) {
        String storeDir = XStartup.getCurrent().getProperty("weather.file.temp.path");
        StringBuilder stringBuilder = new StringBuilder(storeDir)
                .append(DateTime.now().toString("yyyyMMdd"));
        File savePath = new File(stringBuilder.toString());
        if (!savePath.exists()) {
            savePath.mkdirs();
        }
        String filePath = stringBuilder.append(File.separator).append(fileName).toString();
        FileUtil.writeFromStream(in, filePath);
        IoUtil.close(in);
        return filePath;
    }
}
